import os
import glob
from typing import Dict
import numpy as np
import xarray as xr


from satio.collections import (AgERA5Collection,
                               SIGMA0TrainingCollection)
from cropclass.geoloader import GDALWarpLoader


class AgERA5YearlyCollection(AgERA5Collection):

    sensor = 'AgERA5'

    def __init__(self, *args, **kwargs):
        """
        collection csv/dataframe should have the paths
        in column 'path'
        e.g.
        date      |    path
        20200101   .../ewoc-agera5-yearly/2020
        """
        super().__init__(*args, **kwargs)
        self._loader = GDALWarpLoader()

    def filter_dates(self, start_date, end_date):
        start_year = int(start_date[:4])
        end_year = int(end_date[:4]) + 1

        df = self.df[(self.df.date >= f'{start_year}0101')
                     & (self.df.date < f'{end_year}0101')]
        return self._clone(df=df, start_date=start_date, end_date=end_date)

    def get_band_filenames(self, band, resolution=None):
        filenames = self.df.apply(
            lambda x:
            f"{x.path}/AgERA5_{band}_"
            f'{x.date.year}.tif',
            axis=1)
        return filenames.values.tolist()

    def load_timeseries(self,
                        *bands,
                        resolution=100,
                        resampling='cubic',
                        **kwargs):
        return super().load_timeseries(*bands,
                                       resolution=resolution,
                                       resampling=resampling,
                                       **kwargs)


class SIGMA0HRLTrainingCollection(SIGMA0TrainingCollection):
    '''Override parent class to support simultaneous
    ascending/descending file loading into one collection
    NOTE: no way to retrieve original orbit direction so once
    you go this way, there is no turning back!
    '''

    def _load_worldcereal(self,
                          row,
                          bands,
                          mask_and_scale=False) -> Dict:
        nc_files_ASC = glob.glob(
            os.path.join(
                row['path'],
                f'*{self.sensor}*ASCENDING*{self.processing_level}*nc'))
        nc_files_DES = glob.glob(
            os.path.join(
                row['path'],
                f'*{self.sensor}*DESCENDING*{self.processing_level}*nc'))

        xarrs = {}

        for nc_files in [nc_files_ASC, nc_files_DES]:
            if len(nc_files) == 0:
                continue

            avail_bands = dict()

            for nc_file in nc_files:
                ds = xr.open_dataset(nc_file, engine='h5netcdf',
                                     mask_and_scale=mask_and_scale)
                avail_bands.update(dict.fromkeys(ds.keys(), nc_file))
                if 'spatial_ref' in avail_bands.keys():
                    avail_bands.pop('spatial_ref')

            if bands is None:
                bands = list(avail_bands.keys())

            for b in bands:
                if b not in avail_bands:
                    raise ValueError("Given band is not available."
                                     f"Available bands: {avail_bands}")

            for b in bands:
                new_xarr = xr.open_dataset(
                    avail_bands[b], engine='h5netcdf',
                    mask_and_scale=mask_and_scale)[b]
                if b not in xarrs.keys():
                    xarrs[b] = new_xarr
                else:
                    xarrs[b] = xr.concat([xarrs[b], new_xarr], dim='timestamp')

        for b in xarrs.keys():
            if xarrs[b].values.dtype == np.float64:
                # Cast to float32
                xarrs[b].values = xarrs[b].values.astype(np.float32)
            xarrs[b] = xarrs[b].sortby('timestamp')

        return xarrs
